/* Copyright (C) 2021 Goethe-University Frankfurt, Frankfurt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Pascal Raisig, Dominik Smith [committer] */

#ifndef UnpackRich_H
#define UnpackRich_H

#include "CbmRichDigi.h"

#include "Timeslice.hpp"

#include <cstddef>
#include <cstdint>
#include <iomanip>
#include <memory>
#include <sstream>
#include <utility>

namespace cbm::algo
{
  class RichMicrosliceReader;

  enum class RichTdcWordType
  {
    TimeData,
    Header,
    Epoch,
    Trailer,
    Debug,
    Error
  };


  struct RichTdcTimeData {
    uint32_t fCoarse       = 0;  // 11 bits
    uint32_t fIsRisingEdge = 0;  // 1 bit
    uint32_t fFine         = 0;  // 10 bits
    uint32_t fChannel      = 0;  // 7 bits
  };


  /** @struct UnpackRichElinkPar
   ** @author Dominik Smith <d.smith@gsi.de>
   ** @since 26 May 2023
   ** @brief RICH Unpacking parameters for one eLink / ASIC
   **/
  struct UnpackRichElinkPar {
    std::vector<double> fToTshift;  ///< TOT shift for different channels
  };


  /** @struct UnpackRichPar
   ** @author Dominik Smith <d.smith@gsi.de>
   ** @since 26 May 2023
   ** @brief Parameters required for the RICH unpacking (specific to one component)
   **/
  struct UnpackRichPar {
    std::map<uint32_t, UnpackRichElinkPar> fElinkParams = {};  ///< Map TRB address to elink params
  };


  /** @struct UnpackRichMoni
   ** @author Dominik Smith <d.smith@gsi.de>
   ** @since 26 May 2023
   ** @brief Monitoring data for RICH unpacking
   **/
  struct UnpackRichMonitorData {
    uint32_t fNumDebugMessage            = 0;  ///< Received debug messages
    uint32_t fNumErrInvalidFirstMessage  = 0;  ///< First message is not HEADER
    uint32_t fNumErrInvalidSecondMessage = 0;  ///< Second message is not EPOCH
    uint32_t fNumErrInvalidLastMessage   = 0;  ///< Last message is not TRAILER
    uint32_t fNumErrWildHeaderMessage    = 0;  ///< TDC header in invalid position
    uint32_t fNumErrWildTrailerMessage   = 0;  ///< TDC trailer in invalid position
    uint32_t fNumErrTdcErrorWord         = 0;  ///< TDC word of error type
    uint32_t fNumErrChannelOutOfBounds   = 0;  ///< TDC channel out of bounds
    uint32_t fNumErrInvalidHubId         = 0;  ///< "SubSubEvent" has invalid ID
    uint32_t fNumErrInvalidHubSize       = 0;  ///< Premature end of hub block
    uint32_t fNumErrExcessLastWords      = 0;  ///< More than exprected trailing words

    bool HasErrors()
    {
      uint32_t numErrors = fNumDebugMessage + fNumErrInvalidFirstMessage + fNumErrInvalidSecondMessage
                           + fNumErrInvalidLastMessage + fNumErrWildHeaderMessage + fNumErrWildTrailerMessage
                           + fNumErrTdcErrorWord + fNumErrChannelOutOfBounds + fNumErrInvalidHubId
                           + fNumErrInvalidHubSize + fNumErrExcessLastWords;
      return (numErrors > 0 ? true : false);
    }
    std::string print()
    {
      std::stringstream ss;
      ss << "errors " << fNumDebugMessage << " | " << fNumErrInvalidFirstMessage << " | " << fNumErrInvalidSecondMessage
         << " | " << fNumErrInvalidLastMessage << " | " << fNumErrWildHeaderMessage << " | "
         << fNumErrWildTrailerMessage << " | " << fNumErrTdcErrorWord << " | " << fNumErrChannelOutOfBounds << " | "
         << fNumErrInvalidHubId << " | " << fNumErrInvalidHubSize << " | " << fNumErrExcessLastWords << " | ";
      return ss.str();
    }
  };


  class UnpackRich {
  public:
    typedef std::pair<std::vector<CbmRichDigi>, UnpackRichMonitorData> resultType;


    /** @brief Default constructor **/
    UnpackRich() {};


    /** @brief Destructor **/
    ~UnpackRich() {};


    /** @brief Algorithm execution
     ** @param  msContent  Microslice payload
     ** @param  msDescr    Microslice descriptor
     ** @param  tTimeslice Unix start time of timeslice [ns]
     ** @return RICH digi data
     **/
    resultType operator()(const uint8_t* msContent, const fles::MicrosliceDescriptor& msDescr,
                          const uint64_t tTimeslice);

    /** @brief Set the parameter container
     ** @param params Pointer to parameter container
     **/
    void SetParams(std::unique_ptr<UnpackRichPar> params) { fParams = *(std::move(params)); }

  private:
    void ProcessTrbPacket(RichMicrosliceReader& reader);

    void ProcessHubBlock(RichMicrosliceReader& reader);

    int ProcessCtsHeader(RichMicrosliceReader& reader, uint32_t subSubEventSize, uint32_t subSubEventId);

    void ProcessSubSubEvent(RichMicrosliceReader& reader, int nofTimeWords, uint32_t subSubEventId);

    void ProcessTimeDataWord(uint32_t epoch, uint32_t tdcWord, uint32_t subSubEventId,
                             std::vector<double>& raisingTime);

    RichTdcWordType GetTdcWordType(uint32_t tdcWord);

    RichTdcTimeData ProcessTimeData(uint32_t tdcWord);

    uint32_t ProcessEpoch(uint32_t tdcWord);

    uint16_t ProcessHeader(uint32_t tdcWord);

    uint16_t ProcessTrailer(uint32_t tdcWord);

    void WriteOutputDigi(int32_t fpgaID, int32_t channel, double time, double tot);

    double CalculateTime(uint32_t epoch, uint32_t coarse, uint32_t fine);

    int32_t GetPixelUID(int32_t fpgaID, int32_t ch);

    bool CheckMaskedDiRICH(int32_t subSubEventId);

  private:
    UnpackRichPar fParams = {};  ///< Parameter container

    std::vector<int32_t> fMaskedDiRICHes;

    bool fbDoToTCorr = true;  //activates ToT correction from parameter file

    uint64_t fCbmTimeMS;
    uint64_t fCbmTimePacket;

    double fMbsCorr = 0.;
    double fPrevLastCh0ReTime[13];  // 12 DiRICHes chnl0 + 1 CTS chnl0
    uint16_t fCurrentSubSubEvent = 0;

    uint64_t fMsRefTime = 0;

    double fToTMin = -20.;
    double fToTMax = 100.;

    /** @brief Default return vector for Unpack function. */
    std::vector<CbmRichDigi> fOutputVec = {};

    /** @brief Storage of monitoring data. */
    UnpackRichMonitorData fMonitor;

    /**
   * @brief Time offset for the system
   * @todo This should be module and channel dependent and included into the asic parameters
  */
    std::int32_t fSystemTimeOffset = 0;
  };


  class RichMicrosliceReader {
  private:
    const uint8_t* fData = nullptr;
    size_t fSize         = 0;
    size_t fOffset       = 0;  // offset in bytes
    size_t fWordCounter  = 0;
    uint32_t fCurWord    = 0;

  public:
    void SetData(const uint8_t* data, size_t size)
    {
      fData        = data;
      fSize        = size;
      fOffset      = 0;
      fWordCounter = 0;
      fCurWord     = 0;
    }

    const uint8_t* GetData() { return fData; }
    size_t GetSize() { return fSize; }
    size_t GetOffset() { return fOffset; }
    size_t GetWordCounter() { return fWordCounter; }
    uint32_t GetCurWord() { return fCurWord; }

    std::string GetWordAsHexString(uint32_t word)
    {
      std::stringstream stream;
      stream << "0x" << std::setfill('0') << std::setw(sizeof(uint32_t) * 2) << std::hex << word;
      return stream.str();
    }

    uint32_t NextWord()
    {
      uint32_t i = ((uint32_t*) (fData + fOffset))[0];
      //swap bytes
      i = (i >> 24) | ((i << 8) & 0x00FF0000) | ((i >> 8) & 0x0000FF00) | (i << 24);
      fOffset += 4;
      fWordCounter++;
      fCurWord = i;
      return i;
    }

    bool IsNextPadding()
    {
      uint32_t nextWord = ((uint32_t*) (fData + fOffset))[0];
      if (nextWord == 0xffffffff) return true;
      return false;
    }

    bool IsLastSubSubEvent(uint32_t subSubEventSize)
    {
      uint32_t i = ((uint32_t*) (fData + fOffset + 4 * subSubEventSize))[0];
      i          = (i >> 24) | ((i << 8) & 0x00ff0000) | ((i >> 8) & 0x0000ff00) | (i << 24);
      if (i == 0x00015555) return true;
      return false;
    }
  };

} /* namespace cbm::algo */

#endif  // UnpackRich_H
